/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file Node_powsqr.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

#ifndef _CAPD_MAP_NODE_POWSQR_HPP_ 
#define _CAPD_MAP_NODE_POWSQR_HPP_ 

namespace capd{
namespace map{

// ---------------------- constructors ----------------------------------

template<typename ScalarType>
PowNode<ScalarType>::PowNode(Node<ScalarType> *a_left, Node<ScalarType> *a_right, int a_order)
  : Node<ScalarType>(a_order,a_left,a_right)
{
  value = new ScalarType[m_order];
  iterator b=begin(), e=end();
  while(b!=e)
  {
    *b = ScalarType(0);
    ++b;
  }
  if(left) ++(left->m_links);
  if(right) ++(right->m_links);
}

template<typename ScalarType>
SqrtNode<ScalarType>::SqrtNode(Node<ScalarType> *a_left, int a_order)
  : Node<ScalarType>(a_order,a_left,0)
{
  value = new ScalarType[m_order];
  iterator b=begin(), e=end();
  while(b!=e)
  {
    *b = ScalarType(0);
    ++b;
  }
  if(left) ++(left->m_links);
}

// ----------------------- destructors ----------------------------------

template<typename ScalarType>
PowNode<ScalarType>::~PowNode()
{
  if( !( --(left->m_links)) ) 
    delete left;
  if( !( --(right->m_links)) ) 
    delete right;
  delete []value;
}

template<typename ScalarType>
SqrtNode<ScalarType>::~SqrtNode()
{
  if( !( --(left->m_links)) ) 
    delete left;
  delete []value;
}

// ------------------------ operators -----------------------------------

// for intervals only
// specialisation for ScalarType=double
template<typename ScalarType>
ScalarType& PowNode<ScalarType>::operator()(int i)
{

  if(i<=m_maxComputedDerivative)
    return value[i];
  m_maxComputedDerivative = i;


  (*left)(i); 
  (*right)(i);
  int c = toInt((right->value[0]).leftBound());

  if(right->value[0]==ScalarType(1.))
    return value[i] = left->value[i];

  if (i==0)
    return value[i] = (
        ScalarType(c) == right->value[i] ?
          power(left->value[i],c) : power(left->value[i],right->value[i])
        );

  ScalarType result(0.);

  if(!(left->value[0].contains(0.0)))
  {
    for(int j=0;j<i;++j)
      result+= (right->value[0]*ScalarType(i)-(right->value[0]+1)*ScalarType(j)) * value[j] * left->value[i-j];
    return value[i] = result/(ScalarType(i) * left->value[0]);
  }

  if(c<=0)
  {
    std::ostringstream bufor;
    bufor << "function error: can not compute Taylor series for x^(" << c << ") if x contains zero";
    throw(std::runtime_error(bufor.str()));
  }

  ScalarType *temp1 = new ScalarType[i+1];
  ScalarType *temp2 = new ScalarType[i+1];
  int j,p;
  int k;
  for(j=0;j<=i;++j)
    temp1[j]=left->value[j];

  for(k=1;k<c;++k)
  {
    for(p=0;p<=i;++p)
    {
      temp2[p] = ScalarType(0.);
      for(j=0;j<=p;++j)
        temp2[p] += left->value[j]*temp1[p-j];
    }
    for(p=0;p<=i;++p)
      temp1[p]=temp2[p];
  }

  result = temp1[i];
  delete []temp1;
  delete []temp2;

  return value[i] = result;
}

template<typename ScalarType>
ScalarType& SqrtNode<ScalarType>::operator()(int i)
{
  if(i<=m_maxComputedDerivative)
    return value[i];
  m_maxComputedDerivative = i;

  (*left)(i);
  if(!i)
  {
    if(!(left->value[0]>0))
    {
      std::ostringstream bufor;
      bufor << "function error: ScalarType sqrt error: nonpositive argument " << left->value[0];
      throw (std::runtime_error(bufor.str()));
    }
    return value[i] = sqrt(left->value[0]);
  }

  ScalarType result(0);
  for(int j=0;j<i;++j)
    result += ScalarType(0.5) * ScalarType(i-3*j) * value[j] * left->value[i-j];

  return value[i] = result / (ScalarType(i) * left->value[0]);
}

// --------------------------------- setOrder -----------------------------

template<typename ScalarType>
void PowNode<ScalarType>::setOrder(int a_order, ScalarType *a_val)
{
  if(m_order==a_order) return;
  delete [] value;

  m_order = a_order;
  value = new ScalarType[m_order];

  if(left) left->setOrder(m_order,a_val);
  if(right) right->setOrder(m_order,a_val);
}

template<typename ScalarType>
void SqrtNode<ScalarType>::setOrder(int a_order, ScalarType *a_val)
{
  if(m_order==a_order) return;
  delete [] value;

  m_order = a_order;
  value = new ScalarType[m_order];

  if(left) left->setOrder(m_order,a_val);
  if(right) right->setOrder(m_order,a_val);
}

}} // namepsace capd::map

#endif // _CAPD_MAP_NODE_POWSQR_HPP_ 

/// @}
