/// @addtogroup poincare
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file BasicPoincareMap2.hpp
///
/// @author Daniel Wilczak
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2008 by the CAPD Group.
//
// Distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details.

// Created on 29 kwiecień 2008, 00:24



#ifndef _BASIC_POINCARE_MAP2_HPP
#define _BASIC_POINCARE_MAP2_HPP
#include <stdexcept>
#include "capd/factrial.h"
#include "poincare/BasicPoincareMap2.h"
#include "poincare/BasicPoincareMap2_fl.hpp"


namespace capd{
namespace poincare{

// -----------------------------------------------------------------------------------------



/**
 *  Computes dT from derivative of the flow.
 *  T(x) is a first return time for point x
 *
 *  @param Px                value of Poincare map P(x)
 *  @param derivativeOfFlow  defivative of the flow \f$ \partial \varphi(T(x),x)/ \partial x \f$
 */

template <typename DS, typename MapT>
typename BasicPoincareMap2<DS, MapT>::VectorType BasicPoincareMap2<DS, MapT>::computeDT(
      const VectorType& Px,
      const MatrixType& derivativeOfFlow
   )
{
   int j,k;
   VectorType result(m_dim);
   VectorType gradientOnPx = gradient(Px);
   for(j=0;j<m_dim;++j)
      for(k=0;k<m_dim;++k)
         result[j] += gradientOnPx[k] * derivativeOfFlow[k][j];

   result /= - gradientOnPx * dynamicalSystem.getField()(Px);
   return result;
}

// -------------------------------------------------------------------

/**
 * Computes derivative of Poincare Map dP
 *
 *  @param Px                value of Poincare map P(x)
 *  @param derivativeOfFlow  defivative of the flow \f$ \partial \varphi(T(x),x)/ \partial x \f$
 *  @param dT                derivative of T(x) : first return time function 
 */

template <typename DS, typename MapT>
typename BasicPoincareMap2<DS, MapT>::MatrixType BasicPoincareMap2<DS, MapT>::computeDP(
      const VectorType& Px,
      const MatrixType& derivativeOfFlow,
      const VectorType& dT
   )
{
   MatrixType result(m_dim,m_dim);
   VectorType fieldOnPx = dynamicalSystem.getField()(Px);
   int i,j;
   for(i=0;i<m_dim;++i)
      for(j=0;j<m_dim;++j)
         result[i][j] = fieldOnPx[i]*dT[j] + derivativeOfFlow[i][j];
   return result;
}

/**
 * Computes derivative of Poincare Map dP
 *
 *  @param Px                value of Poincare map P(x)
 *  @param derivativeOfFlow  defivative of the flow \f$ \partial \varphi(T(x),x)/ \partial x \f$
 */

template <typename DS, typename MapT>
typename BasicPoincareMap2<DS, MapT>::MatrixType BasicPoincareMap2<DS, MapT>::computeDP(
      const VectorType& Px,
      const MatrixType& derivativeOfFlow
   )
{
  VectorType dT = computeDT(Px, derivativeOfFlow);
  return computeDP(Px, derivativeOfFlow, dT);
}
// -------------------------------------------------------------------

template <typename DS, typename MapT>
typename BasicPoincareMap2<DS, MapT>::MatrixType BasicPoincareMap2<DS, MapT>::computeD2T(
      const VectorType& Px,
      const MatrixType& derivativeOfFlow,
      const VectorType& DT,
      const MatrixType& DP,
      MatrixType hessianOfFlow[]
   )
{
   int i,j,k,s;
   MatrixType D2T(m_dim,m_dim);
   VectorType fieldOnPx = dynamicalSystem.getField()(Px);
   MatrixType derOfVectorFieldOnPx = dynamicalSystem.getField()[Px];
   VectorType gradientOnPx = gradient(Px);

   for(j=0;j<m_dim;++j)
      for(i=0;i<m_dim;++i)
      {
         ScalarType temp1=0,temp2=0;
         for(k=0;k<m_dim;++k)
         {
            D2T[j][i] -= (gradientOnPx[k] * hessianOfFlow[k][j][i]);
            for(s=0;s<m_dim;++s)
            {
               ScalarType temp3 = gradientOnPx[k] * derOfVectorFieldOnPx[k][s];
               temp1 -= temp3 * DP[s][i];
               temp2 -= temp3 * derivativeOfFlow[s][j];
            }
         }
         D2T[j][i] += DT[j] *temp1 + DT[i]*temp2;
      }
   D2T /= (gradientOnPx * fieldOnPx);
   return D2T;
}

// -------------------------------------------------------------------

template <typename DS, typename MapT>
void BasicPoincareMap2<DS, MapT>::computeD2P(
         const VectorType& Px,
         const MatrixType& derivativeOfFlow,
         const VectorType& DT,
         const MatrixType& DP,
         MatrixType hessianOfFlow[],
         const MatrixType& D2T,
         MatrixType result[]
        )
{
   int i,j,k,s;
   MatrixType derOfVectorFieldOnPx = dynamicalSystem.getField()[Px];
   VectorType fieldOnPx = dynamicalSystem.getField()(Px);

   for(i=0;i<m_dim;++i)
      for(j=0;j<m_dim;++j)
         for(k=0;k<m_dim;++k)
         {
            result[i][j][k] = fieldOnPx[i] * D2T[j][k] + hessianOfFlow[i][j][k];
            for(s=0;s<m_dim;++s)
            {
               result[i][j][k] += DT[k] * derOfVectorFieldOnPx[i][s] * derivativeOfFlow[s][j];
               result[i][j][k] += DT[j] * derOfVectorFieldOnPx[i][s] * DP[s][k];
            }
         }
}

// ------------------------------------------------------------------------------------------

/**
 *  Computes a partial derivatives of a Poincare map to an arbitrarily order. 
 *  
 *  @param PhiSeries  partial derivatives of the flow to a given order
 *
 *  @remark Computations are valid only for linear sections.
 */
template <typename DS, typename MapT>
template<typename CnCoeffType>
CnCoeffType BasicPoincareMap2<DS, MapT>::computePartialDerivatives(const CnCoeffType& PhiSeries)
{
  typedef typename CnCoeffType::Multipointer Multipointer;

  const int rank = PhiSeries.rank();
  if(m_dim != PhiSeries.dimension())
      throw std::runtime_error("Incompatible dimensions in BasicPoincareMap2::computePartialDerivatives");
  if(rank > this->dynamicalSystem.getRank())
      throw std::runtime_error("Incompatible orders in BasicPoincareMap2::computePartialDerivatives");

  CnCoeffType PSeries(m_dim,rank); // result
  PSeries() = PhiSeries();
  if(rank<=0) return PSeries;

// fieldOnPx contains a series of vectorField computed at Px
  const typename DS::SeriesContainerType& DtPhiSeries = dynamicalSystem.computeDaPhi(PhiSeries,rank+1);

  VectorType gradientOnPx = gradient(PhiSeries());
  VectorType fieldOnPx =  getField()(PhiSeries());
  ScalarType scalarProduct = - gradientOnPx * fieldOnPx;

// first we compute separately dt/dx
  CnCoeffType TSeries(m_dim,rank);
  int i,j,k,s,r,p;
  for(j=0;j<m_dim;++j)
      for(i=0;i<m_dim;++i)
        TSeries(0,j) += gradientOnPx[i] * PhiSeries(i,j);
  for(j=0;j<m_dim;++j)
      TSeries(0,j) /= scalarProduct;

// and we compute dP/dx
  for(i=0;i<m_dim;++i)
      for(j=0;j<m_dim;++j)
        PSeries(i,j) = fieldOnPx[i]*TSeries(0,j) + PhiSeries(i,j);

// recursive computation of higher order partial derivatives
  for(p=2;p<=rank;++p)
  {
      Multipointer a = PSeries.first(p);
      do
      {
        VectorType temp = PhiSeries(a);
        VectorType temp2(m_dim);

        for(k=2;k<=p;++k)
        {
            const typename Multipointer::IndicesSet& is = Multipointer::generateList(p,k);
            typename Multipointer::IndicesSet::const_iterator b=is.begin(), e=is.end();
            while(b!=e)
            {
              ScalarType product=TSeries( 0,a.subMultipointer((*b)[0]) );
              for(j=1;j<k;++j)
                  product *= TSeries( 0,a.subMultipointer((*b)[j]) );

              DtPhiSeries.takeVector(Multipointer(0),k,temp2);
              temp += (product*factorial(k)) * temp2;

              for(s=0;s<k;++s)
              {
                  ScalarType product2(1.);
                  for(j=0;j<k;++j)
                    if(j!=s)
                        product2 *= TSeries( 0,a.subMultipointer((*b)[j]) );

                  DtPhiSeries.takeVector(a.subMultipointer((*b)[s]),k-1,temp2);
                  temp += product2*factorial(k-1)*temp2;
              }
              ++b;
            }
        }

        TSeries(0,a) = (gradientOnPx*temp)/scalarProduct;
        PSeries(a) = temp + TSeries(0,a)*fieldOnPx;
      }while(PSeries.next(a));
  }

  return PSeries;
}


}} // namespace capd::poincare

#endif // _BASIC_POINCARE_MAP2_HPP

/// @}


