// This is HASHSETS.H. Copyright (C) 2002-2004 by Pawel Pilarczyk.
// It contains templates for an expandable table, a set and a multivalued map.
// This is free software. No warranty. Consult 'license.txt' for details.
// Started in January 2002. Last revision: September 25, 2004.


#ifndef _HASHSETS_H_
#define _HASHSETS_H_

#include "options.h"
#include "textfile.h"
#include "integer.h"

#include <cstdlib>
#include <ctime>
#include <iostream>


// class templates defined within this header file (in this order):
template <class element>
class multitable;
template <class element>
class hashedset;
template <class domelement, class imgelement>
class mvmap;


//Doc//
// --------------------------------------------------
// ------------------ multi table -------------------
// --------------------------------------------------
// A table that is actually built of many smaller tables.
// It extends automatically upon use of elements that are outside its range.

#define DEFAULTPIECESIZE 32

template <class element>
class multitable
{
public:
	// default constructor for a table with the given size
	// of each piece (should be a power of 2 or is rounded up)
	multitable (int piecesize = 0);

	// copying constructor
	multitable (const multitable<element> &m);

	// operator =
	multitable<element> &operator =
		(const multitable<element> &m);

	// default destructor
	~multitable ();

	// take the element's reference for reading/writing
	element &operator [] (int n);

	// take the element for reading only (error if does not exist)
	element &operator () (int n) const;

	// allocate parts of the table for holding 'n' elements;
	// the table is still able to grow if necessary
	void allocate (int n);

	// fill the table up to the given pos. with the given element
	void fill (const element &e, int n);

protected:
	// the number of pieces ready to allocate
	int npieces;

	// the number of bits to shift the index of an element
	// in the table and the mask to get the offset in a piece
	int shiftbits, offsetmask;

	// actual tables
	element **tab;

	// increase the number of pieces to the desired one
	void increase (int n);
}; /* class multitable */
//End//

// --------------------------------------------------

template <class element>
multitable<element>::multitable (int piecesize)
{
	tab = NULL;
	npieces = 0;
	if (piecesize <= 0)
		piecesize = DEFAULTPIECESIZE;
	shiftbits = 1;
	while ((1 << shiftbits) < piecesize)
		shiftbits ++;
	offsetmask = (1 << shiftbits) - 1;
	return;
} /* multitable<element>::multitable */

template <class element>
multitable<element>::multitable (const multitable<element> &m):
	npieces (m. npieces), shiftbits (m. shiftbits),
	offsetmask (m. offsetmask)
{
	int piecesize = 1 << shiftbits;
	tab = new element * [npieces];
	if (!tab)
		throw "Cannot alloc mem in copying constructor of a table.";
	for (int i = 0; i < npieces; i ++)
		if (m. tab [i])
		{
			tab [i] = new element [piecesize];
			if (!tab [i])
				throw "No memory in copying constr.";
			for (int j = 0; j < piecesize; j ++)
				tab [i] [j] = m. tab [i] [j];
		}
		else
			tab [i] = NULL;
	return;
} /* multitable<element>::multitable */

template <class element>
multitable<element> &multitable<element>::operator =
	(const multitable<element> &m)
{
	// have all the tables been deallocated?
	int deallocated = 0;

	// if the size of pieces does not match, they must be deallocated
	if (shiftbits != m. shiftbits)
	{
		// deallocate all the pieces
		for (int i = 0; i < npieces; i ++)
			if (tab [i])
			{
				delete [] tab [i];
				tab [i] = NULL;
			}

		deallocated = 1;
		shiftbits = m. shiftbits;
		offsetmask = m. offsetmask;
	}

	// if the number of pieces is not the same, change the table
	// and for the moment gather in the new table all the pieces
	// that are already allocated
	if (npieces != m. npieces)
	{
		// allocate a new table of pieces
		element **newtab = (m. npieces) ?
			(new element * [m. npieces]) : NULL;
		if (!newtab && m. npieces)
			throw "No memory for a table in operator =.";

		// if there may be some not deallocated elements,
		// gather them at the beginning of the table
		// and set the rest of the pointers to NULLs
		if (!deallocated)
		{
			// 'i' points to the current entry in the new table,
			// 'j' points to the current entry in the old table
			int i = 0, j = 0;
			while (i < m. npieces)
			{
				// find an allocated piece in the old table
				while ((j < npieces) && !tab [j])
					j ++;
				// if found, take it to the new table
				if (j < npieces)
					newtab [i ++] = tab [j ++];
				// otherwise zero the rest of the new table
				else
					while (i < m. npieces)
						newtab [i ++] = NULL;
			}
			// if there are some pieces remaining, delete them
			while (j < npieces)
			{
				if (tab [j])
					delete [] tab [j];
				j ++;
			}
		}
		else
		{
			for (int i = 0; i < m. npieces; i ++)
				newtab [i] = NULL;
		}

		if (tab)
			delete [] tab;
		tab = newtab;
		npieces = m. npieces;
	}

	// if the table is empty, then finish now
	if (!npieces)
		return *this;

	// copy the data from 'm' to the current table;
	// try to use pieces which are already allocated
	int first_nonempty = 0, first_empty = 0, pos = 0;
	int piecesize = 1 << shiftbits;

	// find the first nonempty piece and the first empty one
	while ((first_nonempty < npieces) && !tab [first_nonempty])
		first_nonempty ++;
	while ((first_empty < npieces) && tab [first_empty])
		first_empty ++;

	// copy all the pieces
	while (pos < npieces)
	{
		if (m. tab [pos])
		{
			if (!tab [pos])
			{
				if (first_nonempty < npieces)
				{
					tab [pos] = tab [first_nonempty];
					tab [first_nonempty ++] = NULL;
				}
				else
				{
					tab [pos] = new element [piecesize];
					if (!tab [pos])
						throw "Error in operator =.";
				}
				first_empty ++;
			}
			else
				first_nonempty ++;

			// copy the source piece to this piece
			for (int i = 0; i < (1 << shiftbits); i ++)
				tab [pos] [i] = m. tab [pos] [i];
		}
		else if (tab [pos])
		{
			if (first_empty < npieces)
			{
				tab [first_empty] = tab [pos];
				first_empty ++;
			}
			else
				delete [] tab [pos];
			first_nonempty ++;
			tab [pos] = NULL;
		}
		else
			first_empty ++;

		// move to the next position
		pos ++;

		// update pointers to the first available [non]empty piece
		while ((first_nonempty < npieces) && !tab [first_nonempty])
			first_nonempty ++;
		while ((first_empty < npieces) && tab [first_empty])
			first_empty ++;
	}

	return *this;
} /* multitable<element>::operator = */

template <class element>
multitable<element>::~multitable ()
{
	if (!tab)
		return;
	for (int i = 0; i < npieces; i ++)
		if (tab [i])
			delete [] tab [i];
	delete [] tab;
	return;
} /* multitable<element>::~multitable */

template <class element>
inline element &multitable<element>::operator [] (int n)
{
	if (n < 0)
		throw "Negative index of an element in a table used.";

	// calculate the number of piece of interest
	int piece = n >> shiftbits;

	// increase the number of pieces if necessary
	if (piece >= npieces)
	{
		int newnpieces = 2 * npieces + 1;
		if (newnpieces <= piece)
			newnpieces = piece + 1;
		increase (newnpieces);
	}

	// allocate the piece if necessary
	if (!tab [piece])
	{
		tab [piece] = new element [1 << shiftbits];
		if (!tab [piece])
			throw "Cannot allocate a piece of a table";
	}

	return tab [piece] [n & offsetmask];
} /* multitable<element>::operator [] */

template <class element>
inline element &multitable<element>::operator () (int n) const
{
	if (n < 0)
		throw "Negative index of an element in a table used.";

	// calculate the number of piece of interest
	int piece = n >> shiftbits;

	if ((piece >= npieces) || (!tab [piece]))
		throw "Non-existent table entry requested.";

	return tab [piece] [n & offsetmask];
} /* multitable<element>::operator () */

template <class element>
void multitable<element>::allocate (int n)
{
	if (n <= 0)
		return;
	int piecesize = 1 << shiftbits;
	int necessarypieces = (n + piecesize - 1) / piecesize;

	// allocate more pieces if needed
	if (necessarypieces > npieces)
		increase (necessarypieces);

	// deallocate unnecessary pieces
	for (int i = necessarypieces; i < npieces; i ++)
		if (tab [i])
		{
			delete [] tab [i];
			tab [i] = NULL;
		}
	return;
} /* multitable<element>::allocate */

template <class element>
void multitable<element>::fill (const element &e, int n)
{
	if (n <= 0)
		return;
	int piecesize = 1 << shiftbits;
	int maxpiece = (n + piecesize - 1) / piecesize;
	if (maxpiece > npieces)
		increase (maxpiece);
	for (int piece = 0; piece < maxpiece; piece ++)
	{
		if (!tab [piece])
		{
			tab [piece] = new element [piecesize];
			if (!tab [piece])
				throw "Too little mem for a piece.";

		}
		if ((piece == maxpiece - 1) && (n & offsetmask))
			piecesize = n & offsetmask;
		for (int i = 0; i < piecesize; i ++)
			tab [piece] [i] = e;
	}
	return;
} /* multitable<element>::fill */

template <class element>
void multitable<element>::increase (int n)
{
	if (n <= npieces)
		throw "Trying to increase a multitable incorrectly.";
	element **newtab = new element * [n];
	if (!newtab)
		throw "Cannot increase a table.";
	int i;
	for (i = 0; i < npieces; i ++)
		newtab [i] = tab [i];
	for (i = npieces; i < n; i ++)
		newtab [i] = NULL;
	delete [] tab;
	tab = newtab;
	npieces = n;
	return;
} /* multitable<element>::increase */


//Doc//
// --------------------------------------------------
// ------------------- hash stat --------------------
// --------------------------------------------------
// This is a small class used to gather and display hashing statistics
// for the hashing tables in the class hashedset.

class hashstat
{
public:
	hashstat ();
	time_t creationtime;
	unsigned long hashhits;
	unsigned long hashmisses;
	unsigned long rehashcount;
}; /* class hashstat */
//End//

// --------------------------------------------------

inline hashstat::hashstat ()
{
	time (&creationtime);
	hashhits = 0;
	hashmisses = 0;
	rehashcount = 0;
	return;
} /* hashstat::hashstat */

// --------------------------------------------------

inline std::ostream &operator << (std::ostream &out, const hashstat &s)
{
	if (!s. hashhits)
		return out;
	out << "hashing: " << s. hashhits << " hits, avg " <<
		((s. hashhits + s. hashmisses) / s. hashhits) << "." <<
		((s. hashhits + s. hashmisses) * 10 / s. hashhits) % 10 <<
		" attempts (" << s. rehashcount << " times rehashed)";
	return out;
} /* operator << */

/*
inline outputstream &operator << (outputstream &out, const hashstat &s)
{
	if (!s. hashhits)
		return out;
	out << "hashing: " << s. hashhits << " hits, avg " <<
		((s. hashhits + s. hashmisses) / s. hashhits) << "." <<
		((s. hashhits + s. hashmisses) * 10 / s. hashhits) % 10 <<
		" attempts (" << s. rehashcount << " times rehashed)";
	return out;
} */ /* operator << */

/*
inline std::ostream &operator << (std::ostream &out, const hashstat &s)
{
	outputstream tout (out);
	tout << s;
	return out;
} */ /* operator << */


//Doc//
// --------------------------------------------------
// ------------------- hashed set -------------------
// --------------------------------------------------
// This is a template for a set of objects of the given type.
// Each of the objects should have two functions for generating hashing keys:
// "unsigned hashkey1 ()" and " unsigned hashkey2 ()".
// Note: If you remove elements which are not at the end of the set,
// then the numbers of other elements can change!

#define INITHASHSIZE 137
#define INITTABSIZE 30
#define STARTHASHING 30

template <class element>
class hashedset
{
public:
	// default constructor for creating an empty set of objects;
	// if you expect the set to keep a lot of elements,
	// you may want to set the initial size to something large
	hashedset (int initialsize = 0, int bequiet = 1);

	// copying constructor
	hashedset (const hashedset<element> &s);

	// operator =
	hashedset &operator = (const hashedset<element> &s);

	// default destructor
	~hashedset (void);

	// find the given element and return its number
	// or -1 if not in the set
	int getnumber (const element &e) const;

	// check if the given element is in the set;
	// return: true = Yes, false = No
	bool check (int n) const;
	bool check (const element &e) const;

	// get the element with the given number from the set
	const element &operator [] (int n) const;
	const element &get (int n) const;

	// add the given element (if not in the set, yet);
	// return its number
	int add (const element &e);

	// remove the given element from the set;
	// return: 0 = Ok, -1 = it was not in the set
	int remove (int n);
	int remove (const element &e);

	// add/remove the entire given set
	hashedset<element> &add (const hashedset<element> &s);
	hashedset<element> &remove (const hashedset<element> &s);

	// get the number of elements by simply projecting the set
	// onto an integer number
	operator int () const;

	// hash statistics
	hashstat *stat;

protected:
	// the number of elements stored in the set
	int nelem;

	// the table of these elements
	multitable<element> tab;

	// is the hashing table in use?
	int hashing;

	// the size of the hashing table
	primeint hashsize;

	// the number of cleared entries in the hashing table
	int hashcleared;

	// hashing table: -1 = free entry, -2 = cleared
	multitable<int> hashtable;

	// find the position in the hashing table at which the number
	// of the given element should be (if there is -1 there, then
	// the number's element should be written there if adding);
	// save to 'addposition' the first cleared position in the
	// hashing table if found or set it to -1 otherwise
	int hashfind (const element &e, int *addpos = NULL) const;

	// replace the old hashing table with a new one;
	// a desired size may be given, otherwise an optimal size
	// is determined and the table is made bigger or smaller
	void rehash (int newsize = 0);

}; /* class hashedset */
//End//

// --------------------------------------------------

template <class element>
hashedset<element>::hashedset (int initialsize, int bequiet):
	nelem (0), tab ((initialsize > 0) ? initialsize : INITTABSIZE),
	hashing (0),
	hashsize (initialsize + (initialsize >> 2) + INITHASHSIZE),
	hashcleared (0), hashtable (hashsize)
{
	if (bequiet)
		stat = NULL;
	else
		stat = new hashstat;
	return;
} /* hashedset<element>::hashedset */

template <class element>
hashedset<element>::hashedset (const hashedset<element> &s):
	nelem (s. nelem), tab (s. tab),
	hashing (s. hashing), hashsize (s. hashsize),
	hashcleared (s. hashcleared), hashtable (s. hashtable)
{
	if (s. stat)
		stat = new hashstat;
	else
		stat = NULL;
	return;
} /* hashedset<element>::hashedset */

template <class element>
hashedset<element> &hashedset<element>::operator =
	(const hashedset<element> &s)
{
	if (s. stat)
		stat = new hashstat (*stat);

	nelem = s. nelem;
	tab = s. tab;
	hashing = s. hashing;
	hashsize = s. hashsize;
	hashcleared = s. hashcleared;
	hashtable = s. hashtable;

	return *this;
} /* multitable<element>::operator = */

template <class element>
hashedset<element>::~hashedset ()
{
	if (!stat)
		return;
	sout << "  " << *stat << '\n';
	delete stat;
	stat = NULL;
	return;
} /* hashedset<element>::~hashedset */

template <class element>
int hashedset<element>::getnumber (const element &e) const
{
	if (hashing)
	{
		int pos = hashfind (e);
		return (hashtable (pos));
	}
	else
	{
		for (int i = 0; i < nelem; i ++)
			if (tab (i) == e)
				return i;
		return -1;
	}
} /* hashedset<element>::getnumber */

template <class element>
inline bool hashedset<element>::check (int n) const
{
	return ((n >= 0) && (n < nelem));
} /* hashedset<element>::check */

template <class element>
inline bool hashedset<element>::check (const element &e) const
{
	return (getnumber (e) >= 0);
} /* hashedset<element>::check */


template <class element>
inline const element &hashedset<element>::get (int n) const
{
	if ((n < 0) || (n >= nelem))
		throw "Trying to get an element out of range.";
	return tab (n);
} /* hashedset<element>::get */

template <class element>
inline const element &hashedset<element>::operator [] (int n) const
{
	return get (n);
} /* hashedset<element>::operator [] */

template <class element>
int hashedset<element>::add (const element &e)
{
	if (!hashing)
	{
		for (int i = 0; i < nelem; i ++)
			if (tab (i) == e)
				return i;
		tab [nelem ++] = e;
		if (nelem >= STARTHASHING)
		{
			hashing = 1;
			rehash ();
		}
		return (nelem - 1);
	}

	// rehash if there is very little free space in the hashing table
	if ((int) hashsize - hashcleared < nelem + (nelem >> 1) + 19)
		rehash ();

	// find the position of the element's number in the hashing table
	int addpos = -1;
	int pos = hashfind (e, &addpos);

	// if it alread was in the set, then just return its number
	if (hashtable (pos) >= 0)
		return hashtable (pos);

	// add this element to the set and return its new number
	if (addpos >= 0)
		pos = addpos;
	hashtable [pos] = nelem;
	tab [nelem] = e;
	return nelem ++;
} /* hashedset<element>::add */

template <class element>
int hashedset<element>::remove (const element &e)
{
	if (!hashing)
	{
		for (int i = 0; i < nelem; i ++)
			if (tab (i) == e)
			{
				tab [i] = tab (-- nelem);
				return 0;
			}
		return -1;
	}

	// find the position in the hashing table with this element's number
	int pos = hashfind (e);

	// if it was not there, then finish
	if (hashtable (pos) < 0)
		return -1;
	int n = hashtable (pos);

	// update the hashing table and the number of elements in the set
	hashtable [pos] = -2;
	nelem --;
	hashcleared ++;

	// if it was not the last element in the set, move the last one
	// to the vacated place
	if (n != nelem)
	{
		pos = hashfind (tab (nelem));
		hashtable [pos] = n;
		tab [n] = tab (nelem);
	}

	// if there are very few elements in the table, turn off hashing
	if (nelem < STARTHASHING / 2)
	{
		hashing = 0;
	}

	// if there are very many cleared entries, then rehash
	else if (hashcleared > nelem + 19)
		rehash ();

	return 0;
} /* hashedset<element>::remove */

template <class element>
inline int hashedset<element>::remove (int n)
{
	if (!check (n))
		return -1;
	if (!hashing)
	{
		nelem --;
		if (n != nelem)
			tab [n] = tab (nelem);
		return 0;
	}
	return remove (tab (n));
} /* hashedset<element>::remove */

template <class element>
hashedset<element> &hashedset<element>::add (const hashedset<element> &s)
{
	for (int i = 0; i < (int) s; i ++)
		add (s [i]);
	return *this;
} /* hashedset<element>::add */

template <class element>
hashedset<element> &hashedset<element>::remove (const hashedset<element> &s)
{
	if ((int) *this < (int) s)
	{
		for (int i = 0; i < (int) *this; i ++)
			if (s. check ((*this) [i]))
				this -> remove (i --);
	}
	else
	{
		for (int i = 0; i < (int) s; i ++)
			remove (s [i]);
	}
	return *this;
} /* hashedset<element>::remove */

template <class element>
inline hashedset<element>::operator int () const
{
	return nelem;
} /* hashedset<element>::operator int */

template <class element>
int hashedset<element>::hashfind (const element &e, int *addpos) const
{
	if (!hashing)
		throw "Hashing is turned off.";

	int pos = (int) (e. hashkey1 () % (unsigned) (int) hashsize);
	if (addpos)
		*addpos = -1;

	// start updating hashing statistics
	if (stat)
		stat -> hashhits ++;

	// find the position of the element in the hashing table
	int number, add = 0;
	while ((number = hashtable (pos)) != -1)
	{
		if ((number >= 0) && (e == tab (number)))
			return (pos);
		if (addpos && (*addpos < 0) && (number == -2))
			*addpos = pos;
		if (!add)
			add = (int) (e. hashkey2 () %
				((unsigned) (int) hashsize - 1)) + 1;
		pos += add;
		if (pos >= (int) hashsize)
			pos -= (int) hashsize;
		if (stat)
			stat -> hashmisses ++;
	}

	// return the position located in the hashing table
	return (pos);

} /* hashedset<element>::hashfind */

template <class element>
void hashedset<element>::rehash (int newsize)
{
	if (stat)
		stat -> rehashcount ++;

	// adjust the new size of the hashing table
	if ((newsize < (nelem << 1) + INITHASHSIZE) ||
		(newsize > (nelem << 2) + INITHASHSIZE))
		newsize = (nelem << 1) + nelem + INITHASHSIZE;

	// check if it is not too large for 16-bit programs
	int x = 0xFFFF;
	if ((x < 0) && (newsize >= 16384))
		throw "Set too large for a 16-bit program.";

	// free unused memory if desired
	if (newsize < hashsize)
	{
		multitable<int> empty;
		hashtable = empty;
	}
	
	// set the new value of the hashing table and re-buid it
	hashsize = newsize;
	hashcleared = 0;

	// build the entire hash table from the beginning
	hashtable. fill (-1, hashsize);
	for (int i = 0; i < nelem; i ++)
	{
		int pos = hashfind (tab (i));
		if (hashtable (pos) != -1)
			throw "A repeated element in a set!";
		hashtable [pos] = i;
	}

	return;
} /* hashedset<element>::rehash */

//Doc//
// --------------------------------------------------

#define SMALL_SIZE true
#define LARGE_SIZE (!SMALL_SIZE)

template <class element>
std::ostream &write (std::ostream &out, const hashedset<element> &s, bool size)
//End//
{
	if (size == SMALL_SIZE)
	{
		out << '{';
		for (int i = 0; i < (int) s; i ++)
			out << (i ? " " : "") << s [i];
		out << '}';
	}
	else
	{
		if ((int) s)
			out << "; " << (int) s << (((int) s == 1) ?
				" element." : " elements.") << '\n';
		if (s. stat && s. stat -> hashhits)
			out << ';' << *(s. stat) << '\n';
		for (int i = 0; i < (int) s; i ++)
			out << s [i] << '\n';
	}
	return out;
} /* write */

//Doc//
template <class element>
std::ostream &operator << (std::ostream &out, const hashedset<element> &s)
//End//
{
	return write (out, s, LARGE_SIZE);
} /* operator << */

//Doc//
template <class element>
std::istream &read (std::istream &in, hashedset<element> &s, bool size)
//End//
{
	// ignore all the comments at the beginning of the input stream
	ignorecomments (in);

	// determine the closing parenthesis
	// and read the opening one if applicable
	int closing = EOF;
	if (size == SMALL_SIZE)
	{
		closing = readparenthesis (in);
		if (closing == EOF)
			throw "An opening parenthesis expected in a set.";
		ignorecomments (in);
	}

	// read the elements until the closing parenthesis is found
	while (in. peek () != closing)
	{
		element e;
		in >> e;
	//	if (!in)
	//		throw "Failed to read an element of a set.";
		s. add (e);
		ignorecomments (in);

		// read a comma between the elements if it is there
		if (in. peek () == ',')
		{
			in. get ();
			ignorecomments (in);
		}
	}

	// read the closing parenthesis if any
	if (closing != EOF)
		in. get ();

	return in;
} /* operator >> */

//Doc//
template <class element>
std::istream &operator >> (std::istream &in, hashedset<element> &s)
//End//
{
	return read (in, s, LARGE_SIZE);
} /* operator >> */

//Doc//
template <class element>
inline hashedset<element> &operator += (hashedset<element> &s,
	const hashedset<element> &u)
//End//
{
	s. add (u);
	return s;
} /* operator += */

//Doc//
template <class element>
inline hashedset<element> &operator -= (hashedset<element> &s,
	const hashedset<element> &u)
//End//
{
	s. remove (u);
	return s;
} /* operator -= */


//Doc//
// --------------------------------------------------
// ---------------- multivalued map -----------------
// --------------------------------------------------

template <class domelement, class imgelement>
class mvmap
{
public:
	// default constructor
	mvmap (int bequiet = 1);

	// default destructor
	~mvmap ();

	// retrieve an element from the domain for reading only
	const domelement &get (int n) const;

	// retrieve the domain of the map for reading only
	const hashedset<domelement> &getdomain () const;

	// retrieve the image of an element for reading only
	const hashedset<imgelement> &operator () (int n) const;
	const hashedset<imgelement> &operator ()
		(const domelement &x) const;

	// get the image of an element for writing; if the element
	// is not identified by its number only and it is not
	// in the domain, then it is added and a reference to its
	// empty image is returned
	hashedset<imgelement> &operator [] (int n);
	hashedset<imgelement> &operator [] (const domelement &x);

	// return the number of elements in the domain of the map
	operator int () const;

	// remove an element from the domain of the map;
	// return true if removed, false if was not in the domain
	bool remove (const domelement &x);
	bool remove (int n);
	void remove (const hashedset<domelement> &x);

	// should the map be quiet?
	int quiet;

protected:

	// the domain of the map
	hashedset<domelement> domain;

	// images of cubes from the domain
	multitable<hashedset<imgelement> > images;

}; /* class mvmap */
//End//

// --------------------------------------------------

template <class domelement, class imgelement>
inline mvmap<domelement,imgelement>::mvmap (int bequiet):
	domain (0, bequiet), images ()
{
	return;
} /* mvmap::mvmap */

template <class domelement, class imgelement>
inline mvmap<domelement,imgelement>::~mvmap ()
{
	return;
} /* mvmap::~mvmap */

template <class domelement, class imgelement>
inline const domelement &mvmap<domelement,imgelement>::get (int n) const
{
	if ((n < 0) || (n >= (int) domain))
		throw "Domain element number out of range.";
	return domain [n];
} /* mvmap::get */

template <class domelement, class imgelement>
inline const hashedset<domelement> &
	mvmap<domelement,imgelement>::getdomain () const
{
	return domain;
} /* mvmap::getdomain */

template <class domelement, class imgelement>
inline const hashedset<imgelement>
	&mvmap<domelement,imgelement>::operator () (int n) const
{
	if ((n < 0) || (n >= (int) domain))
		throw "Domain element number out of range.";
	return images (n);
} /* mvmap::operator () */

template <class domelement, class imgelement>
inline const hashedset<imgelement>
	&mvmap<domelement,imgelement>::operator ()
	(const domelement &q) const
{
	int n = domain. getnumber (q);
	if (n < 0)
		throw "Element not in the domain.";
	return images (n);
} /* mvmap::operator () */

template <class domelement, class imgelement>
inline hashedset<imgelement>
	&mvmap<domelement,imgelement>::operator [] (int n)
{
	if ((n < 0) || (n >= (int) domain))
		throw "Domain element number out of range.";
	return images [n];
} /* mvmap::operator [] */

template <class domelement, class imgelement>
inline hashedset<imgelement>
	&mvmap<domelement,imgelement>::operator []
	(const domelement &q)
{
	int n = domain. add (q);
	return images [n];
} /* mvmap::operator [] */

template <class domelement, class imgelement>
inline mvmap<domelement,imgelement>::operator int () const
{
	return (int) domain;
} /* mvmap::operator int */

template <class domelement, class imgelement>
inline bool mvmap<domelement,imgelement>::remove (const domelement &x)
{
	return remove (domain. getnumber (x));
} /* mvmap::remove */

template <class domelement, class imgelement>
inline bool mvmap<domelement,imgelement>::remove (int n)
// WARNING: This procedure uses the specific way elements are removed from
// a hashed set. If that way is changed, this procedure may not work anymore.
{
	if ((n < 0) || (n >= (int) domain))
		return false;
	domain. remove (n);
	images [n] = images [(int) domain];
	hashedset<imgelement> empty;
	images [(int) domain] = empty;
	return true;
} /* mvmap::remove */

template <class domelement, class imgelement>
inline void mvmap<domelement,imgelement>::remove
	(const hashedset<domelement> &x)
{
	for (int i = 0; i < (int) x; i ++)
		remove (x [i]);
	return;
} /* mvmap::remove */

//Doc//
// --------------------------------------------------

template <class domelement, class imgelement>
hashedset<imgelement> &retrieveimage (const mvmap<domelement,imgelement> &m,
	hashedset<imgelement> &img)
// Add images of all the elements from the domain of the map to 'img'.
//End//
{
	for (int i = 0; i < (int) m; i ++)
		img. add (m (i));
	return img;
} /* retrieveimage */

//Doc//
template <class domelement, class imgelement>
hashedset<imgelement> &creategraph (const mvmap<domelement,imgelement> &m,
	hashedset<imgelement> &graph)
// Add a graph of a multivalued map to the given set.
//End//
{
	for (int i = 0; i < (int) m; i ++)
	{
		const domelement &e = m. get (i);
		const hashedset<imgelement> &f = m (i);
		for (int j = 0; j < (int) f; j ++)
			graph. add (e * f [j]);
	}
	return graph;
} /* creategraph */

//Doc//
// --------------------------------------------------

template <class domelement, class imgelement>
std::istream &readdomain (std::istream &in, hashedset<domelement> &dom,
	const mvmap<domelement,imgelement> &)
// Read the domain of a multivalued map.
//End//
{
	ignorecomments (in);
	while (in. peek () != EOF)
	{
		domelement e;
		in >> e;
	//	if (!in)
	//		throw "Failed to read a domain element of a map.";
		dom. add (e);

		// read the map's arrow
		ignorecomments (in);
		while (in. peek () == '-')
			in. get ();
		if (in. peek () == '>')
			in. get ();

		ignorecomments (in);
		int closing = readparenthesis (in);

		ignorecomments (in);
		while (in. peek () != closing)
		{
			imgelement junk;
			in >> junk;
		//	if (!in)
		//		throw "Failed to read an image element.";
			ignorecomments (in);
			if (in. peek () == ',')
			{
				in. get ();
				ignorecomments (in);
			}
		}

		if (closing != EOF)
			in. get ();
		ignorecomments (in);
	}
	return in;
} /* readdomain */

//Doc//
template <class domelement, class imgelement>
std::istream &readimage (std::istream &in, hashedset<imgelement> &img,
	const mvmap<domelement,imgelement> &)
// Read the image of a multivalued map.
//End//
{
	ignorecomments (in);
	while (in. peek () != EOF)
	{
		domelement e;
		in >> e;
	//	if (!in)
	//		throw "Failed to read a domain element of a map.";

		// read the map's arrow
		ignorecomments (in);
		while (in. peek () == '-')
			in. get ();
		if (in. peek () == '>')
			in. get ();

		ignorecomments (in);
		read (in, img, SMALL_SIZE);

		ignorecomments (in);
	}
	return in;
} /* readimage */

//Doc//
template <class domelement, class imgelement>
std::istream &readselective (std::istream &in, mvmap<domelement,imgelement> &m,
	const hashedset<domelement> &dom1, const hashedset<domelement> &dom2)
// Read a restriction of a multivalued map to the union of the given sets.
//End//
{
	if (!(int) dom1 && !(int) dom2)
	{
		sout << "Warning: The domain of the map is empty.\n";
		return in;
	}

	ignorecomments (in);
	while (in. peek () != EOF)
	{
		domelement e;
		in >> e;
	//	if (!in)
	//		throw "Failed to read a domain element of a map.";

		// read the map's arrow
		ignorecomments (in);
		while (in. peek () == '-')
			in. get ();
		if (in. peek () == '>')
			in. get ();

		ignorecomments (in);
		if (dom1. check (e) || dom2. check (e))
			read (in, m [e], SMALL_SIZE);
		else
		{
			int closing = readparenthesis (in);
	
			ignorecomments (in);
			while (in. peek () != closing)
			{
				imgelement junk;
				in >> junk;
			//	if (!in)
			//		throw "Failed to read an img elem.";
				ignorecomments (in);
				if (in. peek () == ',')
				{
					in. get ();
					ignorecomments (in);
				}
			}
	
			if (closing != EOF)
				in. get ();
		}
		ignorecomments (in);
	}
	return in;
} /* readselective */

//Doc//
template <class domelement, class imgelement>
std::istream &readrestriction (std::istream &in, mvmap<domelement,imgelement> &m,
	const hashedset<domelement> &dom, const hashedset<imgelement> &img)
// Read a restriction of a multivalued map to the two given sets.
//End//
{
	if (!(int) dom)
	{
		sout << "Warning: The domain of the map is empty.\n";
		return in;
	}

	ignorecomments (in);
	while (in. peek () != EOF)
	{
		domelement e;
		in >> e;
	//	if (!in)
	//		throw "Failed to read a domain element of a map.";

		// read the map's arrow
		ignorecomments (in);
		while (in. peek () == '-')
			in. get ();
		if (in. peek () == '>')
			in. get ();

		ignorecomments (in);
		if (dom. check (e))
		{
			hashedset<imgelement> &y = m [e];
			hashedset<domelement> x;
			read (in, x, SMALL_SIZE);
			for (int i = 0; i < (int) x; i ++)
				if (img. check (x [i]))
					y. add (x [i]);
		}
		else
		{
			int closing = readparenthesis (in);
	
			ignorecomments (in);
			while (in. peek () != closing)
			{
				imgelement junk;
				in >> junk;
			//	if (!in)
			//		throw "Failed to read an img elem.";
				ignorecomments (in);
				if (in. peek () == ',')
				{
					in. get ();
					ignorecomments (in);
				}
			}
	
			if (closing != EOF)
				in. get ();
		}
		ignorecomments (in);
	}
	return in;
} /* readrestriction */

//Doc//
template <class domelement, class imgelement>
inline std::istream &readselective (std::istream &in, mvmap<domelement,imgelement> &m,
	const hashedset<domelement> &dom)
// Read a restriction of a multivalued map to the given set.
//End//
{
	hashedset<domelement> empty;
	return readselective (in, m, dom, empty);
} /* readselective */


//Doc//
// --------------------------------------------------

template <class domelement, class imgelement>
std::ostream &operator << (std::ostream &out, const mvmap<domelement,imgelement> &m)
//End//
{
	for (int i = 0; i < (int) m; i ++)
	{
		out << m. get (i) << " -> ";
		write (out, m (i), SMALL_SIZE) << '\n';
	}
	return out;
} /* operator << */

//Doc//
template <class domelement, class imgelement>
std::istream &operator >> (std::istream &in, mvmap<domelement,imgelement> &m)
//End//
{
	ignorecomments (in);
	while (in. peek () != EOF)
	{
		domelement e;
		in >> e;
	//	if (!in)
	//		throw "Failed to read a domain element of a map.";

		// read the map's arrow
		ignorecomments (in);
		while (in. peek () == '-')
			in. get ();
		if (in. peek () == '>')
			in. get ();

		ignorecomments (in);
		read (in, m [e], SMALL_SIZE);

		ignorecomments (in);
	}
	return in;
} /* operator >> */

#endif // _HASHSETS_H_


